@@ -75,7 +75,7 @@ require "stringio"
 # 
 class FasterCSV
   # The version of the installed library.
-  VERSION = "1.2.0".freeze
+  VERSION = "1.2.3".freeze
   
   # 
   # A FasterCSV::Row is part Array and part Hash.  It retains an order for the
@@ -712,6 +712,13 @@ class FasterCSV
   # 
   FieldInfo = Struct.new(:index, :line, :header)
   
+  # A Regexp used to find and convert some common Date formats.
+  DateMatcher     = / \A(?: (\w+,?\s+)?\w+\s+\d{1,2},?\s+\d{2,4} |
+                            \d{4}-\d{2}-\d{2} )\z /x
+  # A Regexp used to find and convert some common DateTime formats.
+  DateTimeMatcher =
+    / \A(?: (\w+,?\s+)?\w+\s+\d{1,2}\s+\d{1,2}:\d{1,2}:\d{1,2},?\s+\d{2,4} |
+            \d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2} )\z /x
   # 
   # This Hash holds the built-in converters of FasterCSV that can be accessed by
   # name.  You can select Converters with FasterCSV.convert() or through the
@@ -732,12 +739,16 @@ class FasterCSV
   # To add a combo field, the value should be an Array of names.  Combo fields
   # can be nested with other combo fields.
   # 
-  Converters = { :integer   => lambda { |f| Integer(f)        rescue f },
-                 :float     => lambda { |f| Float(f)          rescue f },
-                 :numeric   => [:integer, :float],
-                 :date      => lambda { |f| Date.parse(f)     rescue f },
-                 :date_time => lambda { |f| DateTime.parse(f) rescue f },
-                 :all       => [:date_time, :numeric] }
+  Converters  = { :integer   => lambda { |f| Integer(f)        rescue f },
+                  :float     => lambda { |f| Float(f)          rescue f },
+                  :numeric   => [:integer, :float],
+                  :date      => lambda { |f|
+                    f =~ DateMatcher ? (Date.parse(f) rescue f) : f
+                  },
+                  :date_time => lambda { |f|
+                    f =~ DateTimeMatcher ? (DateTime.parse(f) rescue f) : f
+                  },
+                  :all       => [:date_time, :numeric] }
 
   # 
   # This Hash holds the built-in header converters of FasterCSV that can be
@@ -768,6 +779,7 @@ class FasterCSV
   # 
   # <b><tt>:col_sep</tt></b>::            <tt>","</tt>
   # <b><tt>:row_sep</tt></b>::            <tt>:auto</tt>
+  # <b><tt>:quote_char</tt></b>::         <tt>'"'</tt>
   # <b><tt>:converters</tt></b>::         +nil+
   # <b><tt>:unconverted_fields</tt></b>:: +nil+
   # <b><tt>:headers</tt></b>::            +false+
@@ -778,6 +790,7 @@ class FasterCSV
   # 
   DEFAULT_OPTIONS = { :col_sep            => ",",
                       :row_sep            => :auto,
+                      :quote_char         => '"', 
                       :converters         => nil,
                       :unconverted_fields => nil,
                       :headers            => false,
@@ -1267,6 +1280,15 @@ class FasterCSV
   #                                       (<tt>$/</tt>) is used.  Obviously,
   #                                       discovery takes a little time.  Set
   #                                       manually if speed is important.
+  # <b><tt>:quote_char</tt></b>::         The character used to quote fields.
+  #                                       This has to be a single character
+  #                                       String.  This is useful for
+  #                                       application that incorrectly use
+  #                                       <tt>'</tt> as the quote character
+  #                                       instead of the correct <tt>"</tt>.
+  #                                       FasterCSV will always consider a
+  #                                       double sequence this character to be
+  #                                       an escaped quote.
   # <b><tt>:converters</tt></b>::         An Array of names from the Converters
   #                                       Hash and/or lambdas that handle custom
   #                                       conversion.  A single converter
@@ -1368,9 +1390,16 @@ class FasterCSV
   # The data source must be open for writing.
   # 
   def <<(row)
-    # handle FasterCSV::Row objects
-    row = row.fields if row.is_a? self.class::Row
-    
+    # Handle FasterCSV::Row objects and Hashes
+    row = case row
+      when self.class::Row then row.fields
+      when Hash            then @headers.map { |header| row[header] }
+      else                      row
+    end
+
+    @headers =  row if header_row?
+    @lineno  += 1
+
     @io << row.map(&@quote).join(@col_sep) + @row_sep  # quote and separate
     
     self  # for chaining
@@ -1533,7 +1562,7 @@ class FasterCSV
             end
           end
         else                  # we found a quoted field...
-          $1.gsub('""', '"')  # unescape contents
+          $1.gsub(@quote_char * 2, @quote_char)  # unescape contents
         end
         ""  # gsub!'s replacement, clear the field
       end
@@ -1582,12 +1611,18 @@ class FasterCSV
   # 
   def init_separators(options)
     # store the selected separators
-    @col_sep = options.delete(:col_sep)
-    @row_sep = options.delete(:row_sep)
+    @col_sep    = options.delete(:col_sep)
+    @row_sep    = options.delete(:row_sep)
+    @quote_char = options.delete(:quote_char)
+
+    if @quote_char.length != 1
+      raise ArgumentError, ":quote_char has to be a single character String"
+    end
     
     # automatically discover row separator when requested
     if @row_sep == :auto
-      if [ARGF, STDIN, STDOUT, STDERR].include? @io
+      if [ARGF, STDIN, STDOUT, STDERR].include?(@io) or
+        (defined?(Zlib) and @io.class == Zlib::GzipWriter)
         @row_sep = $INPUT_RECORD_SEPARATOR
       else
         begin
@@ -1612,7 +1647,14 @@ class FasterCSV
               break
             end
           end
-          @io.seek(saved_pos)  # reset back to the remembered position 
+          # tricky seek() clone to work around GzipReader's lack of seek()
+          @io.rewind
+          # reset back to the remembered position
+          while saved_pos > 1024  # avoid loading a lot of data into memory
+            @io.read(1024)
+            saved_pos -= 1024
+          end
+          @io.read(saved_pos) if saved_pos.nonzero?
         rescue IOError  # stream not opened for reading
           @row_sep = $INPUT_RECORD_SEPARATOR
         end
@@ -1620,8 +1662,13 @@ class FasterCSV
     end
     
     # establish quoting rules
+    do_quote = lambda do |field|
+      @quote_char                                      +
+      String(field).gsub(@quote_char, @quote_char * 2) +
+      @quote_char
+    end
     @quote = if options.delete(:force_quotes)
-      lambda { |field| %Q{"#{String(field).gsub('"', '""')}"} }
+      do_quote
     else
       lambda do |field|
         if field.nil?  # represent +nil+ fields as empty unquoted fields
@@ -1629,8 +1676,9 @@ class FasterCSV
         else
           field = String(field)  # Stringify fields
           # represent empty fields as empty quoted fields
-          if field.empty? or field.count(%Q{\r\n#{@col_sep}"}).nonzero?
-            %Q{"#{field.gsub('"', '""')}"}  # escape quoted fields
+          if field.empty? or
+             field.count("\r\n#{@col_sep}#{@quote_char}").nonzero?
+            do_quote.call(field)
           else
             field  # unquoted field
           end
@@ -1645,19 +1693,24 @@ class FasterCSV
     @skip_blanks = options.delete(:skip_blanks)
     
     # prebuild Regexps for faster parsing
-    @parsers    = {
+    esc_col_sep = Regexp.escape(@col_sep)
+    esc_row_sep = Regexp.escape(@row_sep)
+    esc_quote   = Regexp.escape(@quote_char)
+    @parsers = {
       :leading_fields =>
-        /\A(?:#{Regexp.escape(@col_sep)})+/,     # for empty leading fields
+        /\A(?:#{esc_col_sep})+/,                 # for empty leading fields
       :csv_row        =>
         ### The Primary Parser ###
-        / \G(?:^|#{Regexp.escape(@col_sep)})     # anchor the match
-          (?: "((?>[^"]*)(?>""[^"]*)*)"          # find quoted fields
+        / \G(?:^|#{esc_col_sep})                 # anchor the match
+          (?: #{esc_quote}( (?>[^#{esc_quote}]*) # find quoted fields
+                            (?> #{esc_quote*2}
+                                [^#{esc_quote}]* )* )#{esc_quote}
               |                                  # ... or ...
-              ([^"#{Regexp.escape(@col_sep)}]*)  # unquoted fields
+              ([^#{esc_quote}#{esc_col_sep}]*)   # unquoted fields
               )/x,
         ### End Primary Parser ###
       :line_end       =>
-        /#{Regexp.escape(@row_sep)}\Z/           # safer than chomp!()
+        /#{esc_row_sep}\z/                       # safer than chomp!()
     }
   end
   
